
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '33.0.0rc2',
    'branch_nick': 'master',
    'revision_id': '0290623118d0890e799b848d72f2c1a3b18b2b04',
    'revno': 12663
}
